/* 
   Copyright (C) 2006 Luke Sandell <lasandell@gmail.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "kio_resources.h"
#include <kinstance.h>
#include <kstandarddirs.h>
#include <kurl.h>
#include <qstring.h>
#include <sys/stat.h>
#include <qstringlist.h>
#include <kglobal.h>
#include <qdir.h>
#include <qfileinfo.h>
#include <kdebug.h>
#include <kio/netaccess.h>
#include <kmimetype.h>
#include <kapplication.h>
#include <klocale.h>
#include <dcopref.h>
#include <kmessagebox.h>

using namespace KIO;

// To simplify our implementation, the decodeURL() member returns a ResourceInfo struct given a URL
struct ResourceInfo
{
    QString name;        // The basename of the file or directory
    QString resType;     // Resource type
    QString relPath;     // Path relative to the resource directory
    QString path;        // Physical path of resource (if it exists)
    QString localPath;   // Physical path of save location
    bool exists;         // Does it exist?
    bool isDir;          // Is it a directory?
    bool isLocal;        // Is it (exclusively) local?
};

// Helper functions
static UDSAtom createAtom( uint uds, long lng );
static UDSAtom createAtom( uint uds, const QString & str );
static UDSEntry createDirEntry( const QString & fileName );
static UDSEntry createEntry( const QString & path );
static void appendUnique( QStringList & dst, const QStringList & src );

extern "C"
{
    int kdemain( int argc, char **argv )
    {
        KInstance instance("kio_resources");
        (void) KGlobal::locale();
        ResourcesProtocol slave( argv[2], argv[3] );
        slave.dispatchLoop();
        return 0;
    }
}

ResourcesProtocol::ResourcesProtocol( const QCString & pool, const QCString & app )
        : SlaveBase( "resources", pool, app )
{
    m_dirs = KGlobal::dirs();
}

void ResourcesProtocol::get
    ( const KURL & url )
{
    ResourceInfo s;
    if ( !decodeURL( url, s ) )
        return;

    if ( !s.exists )
        error( ERR_DOES_NOT_EXIST, url.prettyURL() );
    else if ( s.isDir )
        error( ERR_IS_DIRECTORY, url.prettyURL() );
    else
    {
        redirection( KURL( s.path ) );
    }

    finished();
}

void ResourcesProtocol::put( const KURL& url, int permissions, bool overwrite, bool resume )
{
    ResourceInfo s;
    if ( !decodeURL( url, s ) )
        return;

    if ( s.exists && s.isDir )
        error( ERR_IS_DIRECTORY, url.prettyURL() );
    else if ( s.exists && !overwrite )
        error( ERR_FILE_ALREADY_EXIST, url.prettyURL() );
    // Trying to save to the root
    else if ( !s.resType )
        error( ERR_SLAVE_DEFINED, i18n("You cannot save to the root of resources:/") );
    else
    {
        locateLocal(s.resType, s.relPath, true);     // Just to create the dirs
        redirection( KURL( s.localPath ) );
    }

    finished();
}

void ResourcesProtocol::stat( const KURL & url )
{
    ResourceInfo s;
    if ( !decodeURL( url, s ) )
        return;

    if ( !s.exists )
        error( ERR_DOES_NOT_EXIST, url.prettyURL() );
    else if ( s.isDir )
        statEntry( createDirEntry( s.name ) );
    else
        //redirection(s.path);
        statEntry( createEntry( s.path ) );

    finished();
}

void ResourcesProtocol::listDir( const KURL & url )
{
    ResourceInfo s;
    QStringList::Iterator it;
    if ( !decodeURL( url, s ) )
        return;

    if ( ! s.isDir )
    {
        error( ERR_IS_FILE, url.prettyURL() );
    }

    else if ( s.name == "/" )
    {
        QStringList types = m_dirs->allTypes();

        // Show a folder for each resource type
        for ( it = types.begin(); it != types.end(); it++ )
        {
            listEntry( createDirEntry( *it ), false );
        }
    }

    else
    {
        
        // List files
        QString filter = s.relPath.isEmpty() ? QString::null : s.relPath + "/*";
        QStringList files = m_dirs->findAllResources( s.resType, filter, false /* no descend */ , true /* unique */);
        for ( it = files.begin(); it != files.end() ; it++ )
        {
            listEntry( createEntry(*it), false );
        }

        // List subdirs
        QStringList dirs = m_dirs->findDirs( s.resType, s.relPath );
        QStringList subdirs;                               // TODO: In Qt 4 we will use QSet
        QDir dirlist;
        
        // First find the subdirs
        for ( it = dirs.begin(); it != dirs.end(); it++ )
        {
            dirlist.setPath(*it);
            dirlist.setFilter( QDir::Dirs );
            appendUnique( subdirs, dirlist.entryList() );   // TODO: Stupid hack, we really need QSet
        }
        
        // Then list them
        for ( it = subdirs.begin(); it != subdirs.end(); it++ )
        {
            listEntry( createDirEntry( *it ), false );
        }
        
        // And tell ResourcesDirNotify that we have entered the directory
        if (s.relPath)
        {
            DCOPRef ref("kded", "resourcesdirnotify");
            ref.send("enterDir", s.resType, s.relPath);
        }
    }

    listEntry( UDSEntry(), true );
    finished();
}

void ResourcesProtocol::del( const KURL & url, bool isFile )
{
    ResourceInfo s;
    if ( ! decodeURL( url, s ) )
        return;

    // Resource does not exist
    else if ( ! s.exists )
        error(ERR_DOES_NOT_EXIST, url.prettyURL());

    // Want to delete a dir
    if ( isFile && s.isDir )
        error(ERR_IS_DIRECTORY, url.prettyURL() );
    else if ( !isFile && !s.isDir )
        error(ERR_IS_FILE, url.prettyURL());

    // System resource exists
    else if ( !s.isLocal || s.relPath.isEmpty() )
        error(ERR_SLAVE_DEFINED, i18n("%1 cannot be deleted because it is a system resource.").arg(url.prettyURL()) );

    // Try delete
    else if ( !s.isDir && ::unlink( QFile::encodeName(s.path) ) )
        error( ERR_CANNOT_DELETE, url.prettyURL() );

    else if ( s.isDir && ::rmdir( QFile::encodeName(s.path) ) )
        error( ERR_CANNOT_DELETE, url.prettyURL() );

    finished();
}

void ResourcesProtocol::mkdir(const KURL &url, int permissions)
{
    ResourceInfo s;
    if ( !decodeURL( url, s ) )
        return;

    if ( s.exists )
        error( ERR_DIR_ALREADY_EXIST, url.prettyURL() );
    else if ( !s.resType || ::mkdir( QFile::encodeName(s.localPath), permissions ) ) // FIXME: is -1 OK?
        error( ERR_COULD_NOT_MKDIR, url.prettyURL() );

    finished();
}

/* TODO: figure out why this doesn't work */

void ResourcesProtocol::rename( const KURL & src, const KURL & dest, bool overwrite )
{
    ResourceInfo s,d;
    if ( !decodeURL(src, s) || !decodeURL(dest, d) )
        return;

    if ( s.relPath.isEmpty() )
        error( ERR_SLAVE_DEFINED, i18n("You cannot rename a top-level resource folder.") );
    else if ( !s.exists )
        error( ERR_DOES_NOT_EXIST, src.prettyURL() );
    else if ( !s.isLocal )
        error( ERR_SLAVE_DEFINED, i18n("%1 cannot be renamed because it is a system resource.").arg(src.prettyURL() ) );
    else if ( !overwrite && d.exists )
        error( ERR_FILE_ALREADY_EXIST, dest.prettyURL() );
    else if ( ::rename( QFile::encodeName(s.path), QFile::encodeName(d.localPath) ) )
        error( ERR_CANNOT_RENAME, src.prettyURL() );

    finished();
}

bool ResourcesProtocol::decodeURL( const KURL & url, ResourceInfo & s )
{
    QString urlPath = url.path(-1);
    s.exists = true;
    s.isDir = false;

    // Error if empty or relative path
    if ( KURL::isRelativeURL( url.url()) )
    {
        error(ERR_MALFORMED_URL, url.prettyURL() );
        finished();
        return false;
    }

    s.resType = urlPath.section('/', 1, 1);
    s.name = urlPath.section('/', -1);
    s.relPath = urlPath.section('/', 2);

    // If we are browsing the root
    if ( s.resType.isEmpty() )
    {
        s.name =  "/";
        s.isDir = true;
        return true;
    }

    // If we are browsing a top-level resource folder
    if ( s.relPath.isEmpty() )
    {
        if ( m_dirs->allTypes().contains(s.resType) )
        {
            s.isDir = true;
            s.name = s.resType;
            return true;
        }
    }

    // Otherwise
    s.path = locate( s.resType, s.relPath );
    if ( !s.path )
    {
        s.path = locate( s.resType, s.relPath + '/' );
        if ( !s.path )
            s.exists = false;
        else
            s.isDir = true;
    }
    else
    {
        s.exists = true;
        s.isDir = false;
    }

    // Make sure the path exists
    if ( !locate(s.resType, s.relPath.section('/', 0, -2 ) + '/' ) )
    {
        error(ERR_DOES_NOT_EXIST, url.prettyURL() );
        finished();
        return false;
    }

    s.localPath = locateLocal( s.resType, s.relPath + (s.isDir ? "/" : ""), false );

    int count = s.isDir ?  m_dirs->findDirs(s.resType, s.relPath).size() : m_dirs->findAllResources(s.resType, s.relPath ).size();
    s.isLocal = count <= 1 && s.path == s.localPath;

    return true;
}

static UDSEntry createEntry( const QString & path )
{
    UDSEntry entry;
    UDSAtom atom;
    QFileInfo info(path);

    // Add name
    entry.append( createAtom( UDS_NAME, path.section( '/', -1 ) ) );

    // Add mode
    entry.append( createAtom( UDS_FILE_TYPE, info.isDir() ? S_IFDIR : 0 ) );

    // Add size
    entry.append( createAtom( UDS_SIZE, info.size() ) );

    // Add modification time
    entry.append( createAtom( UDS_MODIFICATION_TIME, info.lastModified().toTime_t() ) );
    entry.append( createAtom( UDS_ACCESS_TIME, info.lastRead().toTime_t() ) );
    //entry.append( createAtom( UDS_CREATION_TIME, info.created().toTime_t() ) );

    // Add mimetype
    KMimeType::Ptr pMimetype = KMimeType::findByPath( path, info.isDir() ? S_IFDIR : 0 );
    entry.append( createAtom( UDS_MIME_TYPE, pMimetype->name() ) );

    return entry;
}

static UDSEntry createDirEntry( const QString & fileName )
{
    UDSEntry entry;

    entry.append( createAtom( UDS_FILE_TYPE, S_IFDIR ) );
    entry.append( createAtom( UDS_NAME, fileName ) );

    return entry;
}

static UDSAtom createAtom( uint uds, long lng )
{
    UDSAtom atom;
    atom.m_uds = uds;
    atom.m_long = lng;

    return atom;
}

static UDSAtom createAtom( uint uds, const QString & str )
{
    UDSAtom atom;
    atom.m_uds = uds;
    atom.m_str = str;

    return atom;
}

/** TODO: With Qt 4, the use of QSet will make this silly little function unecessary **/
static void appendUnique( QStringList & dst, const QStringList & src )
{
    QStringList::const_iterator it;
    for ( it = src.begin(); it != src.end(); it++ )
    {
        if ( !dst.contains(*it) )
        {
            dst.append(*it);
        }
    }
}
